-------------------------------------------------------------------
-- Date			: 11 APR 2012
-- Database		: Ms SQL Server
-- Description	: ISAMESSO Cached Wallet Maintenance Script
-- IMS Version	: 8.1.x / 8.2.x
--
--	Syntax		: call sp_pruneCachedWallet '<del mode>','<retain date>',<max user> 
--  * <del mode> - Mandatory, text, case sensitive  
--				 - 'ABOUT' displays script information 
--				 - 'DELETE' performs actual delete 
--				 - 'HELP' displays script syntax and example 
--				 - 'SIMULATE' simulates deletion without actual delete 
--	* <retain date>	- Optional, text, date in YYYYMMDD format
--					- Process only Cached Wallets created before specified retain date
--					- Leave blank to process all Cached Wallets
--	* <max users>  - Mandatory, numeric
--				   - Maximum number of ISAMESSO users to be processed
--				   - Enter 0 to process all the ISAMESSO users
-------------------------------------------------------------------
--
-- EXEC sp_pruneCachedWallet 'SIMULATE','',5000 		- Simulate deletion up to 5K random users' dangling cached wallet
-- EXEC sp_pruneCachedWallet 'DELETE','20110101', 10000 - Delete up to 10K random users' dangling cached wallet created before 01-Jan-2011
-- EXEC sp_pruneCachedWallet 'SIMULATE','',0 			- Simulate deletion of all users' dangling cached wallet without performing actual delete. 
--
-- DROP PROCEDURE sp_pruneCachedWallet
-- EXEC sp_pruneCachedWallet 'ABOUT','',0 
-- EXEC sp_pruneCachedWallet 'HELP','',0 
--
-- Tables affected : 
--  IMSCertificate
--  IMSSociUniqueAttribute
--	IMSSociAttribute
--	IMSSoci
--
-- Changes Log :
--
-- v1.0.3 - Force creation of temp table column's collate to SQL_Latin1_General_CP1_CS_AS 
--			to support SQL Server with different default collate  
--		  - Fixed temp table column name for case sensitive collate
--		
-- v1.0.2 - Added IMS schema check
-- 		  - Change the load of ISAMESSO user working list from IMSIdentity to IMSSoci
-- 		  - Skip randomise ISAMESSO user list process if limit is NOT specified
--        - Added 'ON COMMIT PRESERVE ROWS' for temporary table creation
--		  - Skip temporary table processing for 'SIMULATE' mode
--		  - Combine 2 steps (INSERT & UPDATE) of temporary processing into 1 (INSERT)
--		  - Interface change from walletDeleteLimit to userProcessLimit to allow better 
-- 			estimation of schedule job time 
--
-- v1.0.1 - Skip to synchonized with other database version
-- 
-- EXECUTE AS USER = 'imsdb7907'
-- EXECUTE AS USER = 'dbo'
--
CREATE PROCEDURE sp_pruneCachedWallet
@IN_operationCmd VARCHAR(16),
@IN_retainDate_yyyymmdd VARCHAR(8),
--@var_wallet_del_limit INTEGER
@IN_userProcessLimit INTEGER
AS
BEGIN

	DECLARE @var_info_total_imsid_processed INT; SET @var_info_total_imsid_processed = 0;
	DECLARE @var_info_total_wallet_processed INT; SET @var_info_total_wallet_processed = 0;
	DECLARE @var_info_total_wallet_deleted INT; SET @var_info_total_wallet_deleted = 0;

	DECLARE @imsID_random_startid INT;
    DECLARE @wallet_del_per_batch INT; SET @wallet_del_per_batch = 10000;
	DECLARE @total_unprocessed_imsID INT; SET @total_unprocessed_imsID = 10000;

	DECLARE @wallet_prune_before_yyyymmdd INT;
	DECLARE @wallet_per_batch INT;
	DECLARE @var_info_script_ver VARCHAR(16) ; SET @var_info_script_ver = '1.0.3';
	DECLARE @var_info_db_ver VARCHAR(128);
	
	DECLARE @var_info_start_time DATETIME;
	DECLARE @var_info_end_time DATETIME;
	DECLARE @var_info_elapsed_time VARCHAR(32);
	
	DECLARE @var_showSyntax BIT; SET @var_showSyntax = 0;
	DECLARE @var_invalid_schema BIT; SET @var_invalid_schema = 0;
	
	DECLARE @var_strSQL NVARCHAR(512);
	DECLARE @var_strSQL2 NVARCHAR(512);
	DECLARE @var_iTmpNum INT;
	DECLARE @var_tableschema NVARCHAR(32);

	DECLARE @var_ims_schema NVARCHAR(32);
	
	
	--Performance Tuning Variable
	DECLARE @var_imsid_per_batch INT; SET @var_imsid_per_batch = 200;
	DECLARE @var_max_del_per_batch INT; SET @var_max_del_per_batch = 10000;
	--Exclude wallet from deletion in specified days
	DECLARE @var_wallet_retain_days INT; SET @var_wallet_retain_days = 30;
	--Cached wallet delete limit, 0 = no limit
	DECLARE @var_wallet_del_limit INT; SET @var_wallet_del_limit = 0;
	--IMS max user, 0
	DECLARE @var_user_process_limit INT; SET @var_user_process_limit = 0;
	-- Randomise user list before processing, 1:enabled 0:disabled
	DECLARE @var_is_randomise_user_list INT; SET @var_is_randomise_user_list = 1;	
	

	-- Initialise var
	SET @var_ims_schema = SCHEMA_NAME();
	SET @var_tableschema = @var_ims_schema + '.';
	SET @var_user_process_limit = @IN_userProcessLimit;
	
	
	-- Verify IMS schema for IMSSoci tables
	DECLARE @var_iSociTableCount INT;
	SET @var_iSociTableCount = (SELECT COUNT(*) FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = @var_ims_schema AND UPPER(TABLE_NAME) like 'IMSSOCI%');
	IF(@var_iSociTableCount<3)
		SET @var_invalid_schema = 1;
			
	IF @var_invalid_schema = 0
	BEGIN

		
		IF @IN_retainDate_yyyymmdd = ''
		BEGIN
			-- Set it to year 3000 (no retain)  
			SET @wallet_prune_before_yyyymmdd = 30000000;
		END ELSE
		BEGIN
			-- validate date format
			IF ISNUMERIC(@IN_retainDate_yyyymmdd)=1
			BEGIN
				SET @var_strSQL = CAST(@IN_retainDate_yyyymmdd AS DATETIME);
				IF @@ERROR>0
				BEGIN
					PRINT N' Error : Invalid retain date - '''+@IN_retainDate_yyyymmdd+'''';
					PRINT N' ';
					SET @var_showSyntax = 1;
				END ELSE
				BEGIN
					-- if valid format get the integer value of yyyyMMdd
					SET @wallet_prune_before_yyyymmdd = CONVERT(INT,@IN_retainDate_yyyymmdd);	
				END;	
		    END ELSE
			BEGIN
				PRINT N' Error : Invalid retain date - '''+@IN_retainDate_yyyymmdd+'''';
				PRINT N' ';
				SET @var_showSyntax = 1;
			END;
		END;


		IF @IN_operationCmd IN ('DELETE','SIMULATE') AND @var_showSyntax = 0
		BEGIN
			SET NOCOUNT ON;
			
  			-- Capture start time
			SET @var_info_start_time = CURRENT_TIMESTAMP;

			-- create temp table
			CREATE TABLE #tmpTblCachedWalletimsID 
			(
				rowID INTEGER NOT NULL,
				imsID VARCHAR (128) COLLATE SQL_Latin1_General_CP1_CS_AS NOT NULL,
				PRIMARY KEY (rowID)
  			);
		  	
  			CREATE TABLE #tmpTblCachedWalletData
			(
				rowID INTEGER NOT NULL,
				imsID VARCHAR (128) COLLATE SQL_Latin1_General_CP1_CS_AS NOT NULL,
				creationHost VARCHAR (128) COLLATE SQL_Latin1_General_CP1_CS_AS NOT NULL,
				creationYYYYMMDD INTEGER NOT NULL,
				creationDays INTEGER DEFAULT 0,
				sociID VARCHAR (128) COLLATE SQL_Latin1_General_CP1_CS_AS NOT NULL,
				PRIMARY KEY (rowID)
			);


			-- Populate the IMS ID temp table from IMSSoci
			SET @var_strSQL =  'INSERT INTO #tmpTblCachedWalletimsID (rowID, imsID) SELECT ';
			
			--limit the user list
			IF @var_user_process_limit>0
			BEGIN
				SET @var_strSQL = @var_strSQL + ' TOP '+CAST(@var_user_process_limit AS NVARCHAR);
			END 
			
			--randomise
			IF (@var_is_randomise_user_list=1)
			BEGIN
				SET @var_strSQL = @var_strSQL + ' ROW_NUMBER() OVER (ORDER BY randID) rowID, imsID FROM 
  					(SELECT imsID, RAND(CHECKSUM(NEWID())) randID FROM IMSSoci where sociID like ''200%'' GROUP BY imsID) TBL'
  			END ELSE
  			BEGIN
				SET @var_strSQL = @var_strSQL + ' ROW_NUMBER() OVER (ORDER BY imsID) rowID, imsID FROM IMSSoci where sociID like ''200%'' GROUP BY imsID';
  			END;
  			EXECUTE(@var_strSQL);

			-- Get the total number of unprocessed imsID 
			SELECT @total_unprocessed_imsID = COUNT(imsID) FROM #tmpTblCachedWalletimsID;
			
			
			-- ###[ WHILE #tmpTblCachedWalletimsID IS NOT EMPTY) ]###--
			WHILE (@total_unprocessed_imsID > 0)
			BEGIN

				-- check if batch number is greater than outstanding imsid
				IF @var_imsid_per_batch > @total_unprocessed_imsID
				BEGIN
					SET @var_imsid_per_batch = @total_unprocessed_imsID;
				END;
		
				-- calc total wallet processed
				SELECT @wallet_per_batch = COUNT(sociID) FROM IMSSoci 
				WHERE imsID IN
				(
				   SELECT imsID FROM #tmpTblCachedWalletimsID WHERE rowID < ((SELECT MIN(rowID) FROM #tmpTblCachedWalletimsID) + @var_imsid_per_batch)
				)
				AND sociID like '200%';							
				
				
				IF @IN_operationCmd = 'DELETE'
				BEGIN

					-- Remove previous batch cached wallet data
					DELETE FROM #tmpTblCachedWalletData;
	
					-- Retrieve current batch of IMS user's cached wallet & calc age of wallet
					INSERT INTO #tmpTblCachedWalletData (rowID, imsID, sociID, creationHost, creationYYYYMMDD, creationDays )
					SELECT 
						ROW_NUMBER() OVER (ORDER BY imsID),
						imsID, sociID, CREATIONHOST, CREATIONYYYYMMDD, CREATIONDAYS
					FROM
					(
						SELECT 
							   S.imsID
							   ,COALESCE(SA1.attrValue,'-') CREATIONHOST
							   ,COALESCE(SA2.attrValue,'0') CREATIONTIME
							   ,CAST(SUBSTRING(SA2.attrValue,1,8) AS INT) CREATIONYYYYMMDD
							   ,DATEDIFF (day, SUBSTRING(SA2.attrValue,1,8),  GETDATE())  CREATIONDAYS 
							   ,S.sociID
						FROM 
						  ( 
						  	--SELECT DISTINCT TOP 200 imsID FROM IMSSoci WHERE sociID like '200%'
						  	SELECT imsID FROM #tmpTblCachedWalletimsID WHERE rowID < ((SELECT MIN(rowID) FROM #tmpTblCachedWalletimsID) + @var_imsid_per_batch )
						  ) T
					   	  INNER JOIN IMSSoci S ON (S.imsID = T.imsID AND S.sociID like '200%') 
						  INNER JOIN IMSSociAttribute SA1 ON (SA1.sociID = S.sociID AND SA1.attrName = 'keyCreationHost') 
						  INNER JOIN IMSSociAttribute SA2 ON (SA2.sociID = S.sociID AND SA2.attrName = 'keyCreationTime')
					) TBL
					WHERE 
				  	(imsID + CREATIONHOST + CREATIONTIME) NOT IN
				  	(
						SELECT 
							S.imsID
							+ COALESCE(SA1.attrValue,'-') 			
							+ MAX(COALESCE(SA2.attrValue,'0')) 		
						FROM 
						  ( 
						  	--SELECT DISTINCT TOP 200 imsID FROM IMSSoci WHERE sociID like '200%' 
						  	SELECT imsID FROM #tmpTblCachedWalletimsID WHERE rowID < ((SELECT MIN(rowID) FROM #tmpTblCachedWalletimsID) + @var_imsid_per_batch )
						  ) T
					   	  INNER JOIN IMSSoci S ON (S.imsID = T.imsID AND S.sociID like '200%') 
						  INNER JOIN IMSSociAttribute SA1 ON (SA1.sociID = S.sociID AND SA1.attrName = 'keyCreationHost') 
						  INNER JOIN IMSSociAttribute SA2 ON (SA2.sociID = S.sociID AND SA2.attrName = 'keyCreationTime')
					  	GROUP BY
					  		S.imsID, 
					  		SA1.attrValue
				  	)
					AND CREATIONDAYS > @var_wallet_retain_days
					AND CREATIONYYYYMMDD < @wallet_prune_before_yyyymmdd;
	
	
					-- Count cached wallet mark for delete 
					SELECT  @wallet_del_per_batch = COUNT(sociID)
						FROM #tmpTblCachedWalletData;
	
					-- If deletion limit is specified
					IF @var_wallet_del_limit > 0 
					BEGIN
						-- Recalc the delete limit if more than allowed deletion
						IF @wallet_del_per_batch>= @var_wallet_del_limit 
						BEGIN
							SET @wallet_del_per_batch= @var_wallet_del_limit;
							SET @total_unprocessed_imsID = 0; -- force exit 		
						END;
						
						SET @var_wallet_del_limit = @var_wallet_del_limit - @wallet_del_per_batch;
					END;
					
					SET @var_iTmpNum = 1;
					WHILE @var_iTmpNum > 0 
					BEGIN
						-- Delete cached wallet data
						BEGIN TRAN
							SET @var_strSQL = N'DELETE TOP('+ CAST(@var_max_del_per_batch AS NVARCHAR) +') FROM '+@var_tableschema;
							SET @var_strSQL2 = N'WHERE sociID IN ( SELECT sociID FROM #tmpTblCachedWalletData WHERE rowID > 0 AND rowID <= '+CAST(@wallet_del_per_batch AS NVARCHAR)+' )';
							EXECUTE( @var_strSQL + N'IMSCertificate ' + @var_strSQL2);
							EXECUTE( @var_strSQL + N'IMSSociUniqueAttribute ' + @var_strSQL2);
							EXECUTE( @var_strSQL + N'IMSSociAttribute ' + @var_strSQL2);
							EXECUTE( @var_strSQL + N'IMSSoci ' + @var_strSQL2);
							SET @var_iTmpNum = @@ROWCOUNT;
						COMMIT TRAN;
					END;					

				END 
				ELSE IF @IN_operationCmd = 'SIMULATE'
				BEGIN
					
					SELECT @wallet_del_per_batch = COUNT(sociID)
					FROM
					(
						SELECT 
							   S.imsID
							   ,COALESCE(SA1.attrValue,'-') CREATIONHOST
							   ,COALESCE(SA2.attrValue,'0') CREATIONTIME
							   ,CAST(SUBSTRING(SA2.attrValue,1,8) AS INT) CREATIONYYYYMMDD
							   ,DATEDIFF (day, SUBSTRING(SA2.attrValue,1,8), GETDATE())  CREATIONDAYS 
							   ,S.sociID
						FROM 
						  ( 
						  	--SELECT DISTINCT TOP 200 imsID FROM IMSSoci WHERE sociID like '200%' 
						  	SELECT imsID FROM #tmpTblCachedWalletimsID WHERE rowID < ((SELECT MIN(rowID) FROM #tmpTblCachedWalletimsID) + @var_imsid_per_batch )
						  ) T
					   	  INNER JOIN IMSSoci S ON (S.imsID = T.imsID AND S.sociID like '200%') 
						  INNER JOIN IMSSociAttribute SA1 ON (SA1.sociID = S.sociID AND SA1.attrName = 'keyCreationHost') 
						  INNER JOIN IMSSociAttribute SA2 ON (SA2.sociID = S.sociID AND SA2.attrName = 'keyCreationTime')
					) TBL
					WHERE 
				  	(imsID + CREATIONHOST + CREATIONTIME) NOT IN
				  	(
						SELECT 
							S.imsID
							+ COALESCE(SA1.attrValue,'-') 			
							+ MAX(COALESCE(SA2.attrValue,'0')) 		
						FROM 
						  ( 
						  	--SELECT DISTINCT TOP 200 imsID FROM IMSSoci WHERE sociID like '200%'
						  	SELECT imsID FROM #tmpTblCachedWalletimsID WHERE rowID < ((SELECT MIN(rowID) FROM #tmpTblCachedWalletimsID) + @var_imsid_per_batch )
						  ) T
					   	  INNER JOIN IMSSoci S ON (S.imsID = T.imsID AND S.sociID like '200%') 
						  INNER JOIN IMSSociAttribute SA1 ON (SA1.sociID = S.sociID AND SA1.attrName = 'keyCreationHost') 
						  INNER JOIN IMSSociAttribute SA2 ON (SA2.sociID = S.sociID AND SA2.attrName = 'keyCreationTime')
					  	GROUP BY
					  		S.imsID, 
					  		SA1.attrValue
				  	)
					AND CREATIONDAYS > @var_wallet_retain_days
					AND CREATIONYYYYMMDD < @wallet_prune_before_yyyymmdd;
					
	
				END;
				
				SET @var_info_total_wallet_processed = @var_info_total_wallet_processed + @wallet_per_batch;
				SET @var_info_total_wallet_deleted = @var_info_total_wallet_deleted + @wallet_del_per_batch;
				SET @var_info_total_imsid_processed = @var_info_total_imsid_processed + @var_imsid_per_batch;
				SET @total_unprocessed_imsID = @total_unprocessed_imsID - @var_imsid_per_batch;
				
				-- Delete current batch of IMS ID
				DELETE FROM #tmpTblCachedWalletimsID WHERE rowID < ((SELECT MIN(rowID) FROM #tmpTblCachedWalletimsID) + @var_imsid_per_batch);
			END;


			-- Capture end time
			SET @var_info_end_time = CURRENT_TIMESTAMP;
			DECLARE @var_timediff DATETIME; SET @var_timediff = @var_info_end_time-@var_info_start_time;
			SET @var_info_elapsed_time = CAST(DATEPART(day,@var_timediff)-1 AS VARCHAR)+'d ' 
				+ CAST(DATEPART(hour,@var_timediff) AS VARCHAR) +'h ' + CAST(DATEPART(minute,@var_timediff) AS VARCHAR)+'m '
				+ CAST(DATEPART(second,@var_timediff) AS VARCHAR)+'s ' + CAST(DATEPART(millisecond,@var_timediff) AS VARCHAR)+'ms ';

			PRINT N'command : sp_pruneCachedWallet ('''+ @IN_operationCmd + ''',''' + @IN_retainDate_yyyymmdd + ''',' + CAST(@var_user_process_limit AS NVARCHAR) + ')';
			PRINT N' Input > Schema : ' + @var_ims_schema;
			PRINT N' Input > Delete mode : ' + @IN_operationCmd;
			
			IF @IN_retainDate_yyyymmdd = ''
				PRINT N' Input  > Retain date : not specified';
			ELSE
				PRINT N' Input  > Retain date : ' +@IN_retainDate_yyyymmdd;
			
			IF @var_user_process_limit = 0
				PRINT N' Input  > Maximum no. of users : no limit';
			ELSE
				PRINT N' Input  > Maximum no. of users : ' + CAST(@var_user_process_limit AS NVARCHAR);		

			
			PRINT N' Output > Start time : ' + convert(nvarchar, @var_info_start_time, 106) + ' ' + convert(nvarchar, @var_info_start_time, 14);
			PRINT N' Output > End time : ' + convert(nvarchar, @var_info_end_time, 106) + ' ' + convert(nvarchar, @var_info_end_time, 14);
			PRINT N' Output > Total elapsed time : ' + @var_info_elapsed_time;
			
			PRINT N' Output > Total ISAMESSO user(s) processed : ' + CAST(@var_info_total_imsid_processed AS NVARCHAR);
			PRINT N' Output > Total Cached Wallet(s) processed : ' + CAST(@var_info_total_wallet_processed AS NVARCHAR);
			
			IF @IN_operationCmd = 'DELETE'
				PRINT N' Output > Total Cached Wallet(s) deleted : ' + CAST(@var_info_total_wallet_deleted AS NVARCHAR);
			ELSE
				PRINT N' Output > Total Cached Wallet(s) pending deletion : ' + CAST(@var_info_total_wallet_deleted AS NVARCHAR);
				
			PRINT N' ';
			

			DROP TABLE #tmpTblCachedWalletimsID;
			DROP TABLE #tmpTblCachedWalletData;

			SET NOCOUNT OFF;
		END	
		ELSE IF @IN_operationCmd = 'ABOUT'
		BEGIN
			SET @var_info_db_ver = REPLACE(REPLACE(@@VERSION,CHAR(10),' '),'	','');
			PRINT N' ISAMESSO Cached Wallet Maintenance Script';
			PRINT N'  - Database Name 	 : ' + DB_NAME();
			PRINT N'  - Schema Name 	 : ' + @var_ims_schema;
			PRINT N'  - Script Version   : ' + @var_info_script_ver;
			PRINT N'  - Database Version : ' + @var_info_db_ver;

		END ELSE 
		BEGIN	
			IF @IN_operationCmd <> 'HELP' AND @var_showSyntax <> 1
			BEGIN
				PRINT N' Error : Invalid delete mode '''+@IN_operationCmd+'''';
				PRINT N' ';
			END;
			SET @var_showSyntax = 1;
		END;
		

	END ELSE
	BEGIN
		PRINT N' Error : Invalid ISAMESSO schema '''+@var_ims_schema+'''';
		PRINT N' ';
		SET @var_showSyntax = 1;
	END;

	IF @var_showSyntax = 1
	BEGIN

		PRINT N' Syntax : sp_pruneCachedWallet ''<del mode>'', ''<retain date>'', <max users> ';	
		PRINT N' * <del mode> - Mandatory, text, case sensitive ';
		PRINT N'  			- ''ABOUT'' displays script information';
		PRINT N'  			- ''DELETE'' performs actual delete';
		PRINT N'  			- ''HELP'' displays script syntax and example';
		PRINT N'  			- ''SIMULATE'' simulates deletion without actual delete';
		
		PRINT N' * <retain date> - Optional, text, date in YYYYMMDD format';
		PRINT N'  				 - Process only Cached Wallets created before the retain date';
		PRINT N'  				 - Leave blank to process all Cached Wallets';
			   
		PRINT N' * <max users> - Optional, numeric';
		PRINT N'  					- Maximum number of ISAMESSO users to be processed';
		PRINT N'  					- Enter 0 to process all the ISAMESSO users';
		
		PRINT N'Example : ';
		PRINT N' * sp_pruneCachedWallet ''SIMULATE'','''',5000  - Simulate deletion up to 5K random users'' dangling Cached Wallet';
		PRINT N' * sp_pruneCachedWallet ''DELETE'',''20110101'',10000  - Delete up to 10K random users'' dangling Cached Wallet created before 01-Jan-2011';
		PRINT N' * sp_pruneCachedWallet ''SIMULATE'','''',0  - Simulate deletion of all users'' dangling Cached Wallet without performing actual delete';		
		
	END;
	
END
